/********************************************************************
	created:	2004/02/12
	filename: 	timer\looper.hpp
	author:		Matt Hurd
	
	purpose:	
*********************************************************************/

#ifndef LOOPER_HPP_2004212
#define LOOPER_HPP_2004212

#include "timer.hpp"
#include <boost/function.hpp>
#include <iostream>

#define MAX_FN_LOOP  1e5

namespace matt {
	inline double __fastcall do_sum()
	{
		static double sum;
		static double i;

		sum = 0.0;
		for (i = 0.0; i < MAX_FN_LOOP ; ++i)
		{
			sum += i * i;
		}
		return sum;
	}
	
	struct looper 
	{
		void do_intrusive( boost::function<double (matt::timer & )> fn, size_t num_trials = 100, bool use_high_priority = true, size_t milliseconds_of_pause = 100 )
		{

			double now;
			timer trial;
			std::vector<double> timing;
			timing.resize(num_trials);

			if (use_high_priority)	{
				SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
			}

			Sleep(DWORD(milliseconds_of_pause));

			for (size_t j = 0; j< num_trials; ++j )
			{

				now = fn( trial );
				timing[j] = now - (overhead_estimate_? *overhead_estimate_ : 0.0);

			}

			std::sort(timing.begin(), timing.end());
			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "                    looper invasive timing estimate \n";
			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "median time = " << time_text(timing[num_trials/2] ) << "\n";
			std::cout << "90% range size = " << time_text(timing[num_trials*95/100] - timing[num_trials / 20])  << "\n";
			std::cout << "widest range size (max - min) = " << time_text(timing[num_trials-1] - timing[0]) << "\n";
			std::cout << "minimum time = " << time_text(timing[0]) << "\n";
			std::cout << "maximum time = " << time_text(timing[num_trials-1]) << "\n";
			std::cout << "50% range = (" << time_text(timing[num_trials/4]) << ", " << time_text(timing[num_trials*3 /4]) << ")\n";
			std::cout << "50% range size = " << time_text(timing[num_trials*3/4] - timing[num_trials /4]) << "\n";

			if (use_high_priority)	{
				SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
			}

		}

		void do_timings( boost::function<void (void)> fn, size_t num_trials = 100, bool use_high_priority = true, size_t milliseconds_of_pause = 100 )
		{

			double now;
			timer trial;
			std::vector<double> timing;
			timing.resize(num_trials);

			if (use_high_priority)	{
				SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
			}

			Sleep(DWORD(milliseconds_of_pause));

			for (size_t j = 0; j< num_trials; ++j )
			{

				trial.restart();
				fn();
				now = trial.elapsed();

				timing[j] = now - (overhead_estimate_? *overhead_estimate_ : 0.0);

			}

			std::sort(timing.begin(), timing.end());
			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "                    looper external timing estimate \n";
			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "median time = " << time_text(timing[num_trials/2] ) << "\n";
			std::cout << "90% range size = " << time_text(timing[num_trials*95/100] - timing[num_trials / 20])  << "\n";
			std::cout << "widest range size (max - min) = " << time_text(timing[num_trials-1] - timing[0]) << "\n";
			std::cout << "minimum time = " << time_text(timing[0]) << "\n";
			std::cout << "maximum time = " << time_text(timing[num_trials-1]) << "\n";
			std::cout << "50% range = (" << time_text(timing[num_trials/4]) << ", " << time_text(timing[num_trials*3 /4]) << ")\n";
			std::cout << "50% range size = " << time_text(timing[num_trials*3/4] - timing[num_trials /4]) << "\n";

			if (use_high_priority)	{
				SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
			}

		}


		static double not_empty()
		{
			static double sum;
			static double i;

			sum = 0.0;
			for (i = 0.0; i < MAX_FN_LOOP ; ++i)
			{
				sum += i * i;
			}

			return sum;

		}

		double estimate_boost_function_overhead( size_t num_trials = 10001, boost::function< double (void)> fn = &looper::not_empty )
		{
			double now;
			timer trial;
			std::vector<double> timing;
			std::vector<double> timing1;
			std::vector<double> timing2;
			timing.resize(num_trials);
			timing1.resize(num_trials);
			timing2.resize(num_trials);
			double answer1 = 0.0;

			SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);

			Sleep(100);

			for (size_t j = 0; j< num_trials; ++j )
			{
				trial.restart();

				answer1 += fn();

				now = trial.elapsed();

				timing1[j] = now;

			}

			double answer2 = 0.0;

			for (size_t j = 0; j< num_trials; ++j )
			{
				trial.restart();

				static double sum;
				static double i;
				sum = 0;
				for (i = 0; i < MAX_FN_LOOP ; ++i)
				{
					sum += i * i;
				}

				answer2 += sum;

				now = trial.elapsed();
				timing2[j] = now;

			}


			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "             estimating indirect function call overhead \n";
			std::cout << "---------------------------------------------------------------------\n";

			std::sort(timing1.begin(), timing1.end());
			std::sort(timing2.begin(), timing2.end());

			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "              boost function measurements \n";
			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "median time = " << timing1[num_trials/2] * 1e6 << " microseconds\n";
			std::cout << "90% range size = " << (timing1[num_trials*95/100] - timing1[num_trials / 20]) * 1e9 << " nanoseconds\n";
			std::cout << "widest range size (max - min) = " << (timing1[num_trials-1] - timing1[0]) * 1e9 << " nanoseconds\n";
			std::cout << "minimum time = " << timing1[0] * 1e6 << " microseconds\n";
			std::cout << "maximum time = " << timing1[num_trials-1] * 1e6 << " microseconds\n";
			std::cout << "50% range = (" << timing1[num_trials/4] * 1e6 << ", " << timing1[num_trials*3 /4] * 1e6 << ") microseconds\n";
			std::cout << "50% range size = " << (timing1[num_trials*3/4] - timing1[num_trials /4]) * 1e9 << " nanoseconds\n";
			std::cout << "stop the optimiser eliding garbage = " << answer1 << "\n";
			std::cout << "stop the optimiser eliding garbage = " << answer2 << "\n";


			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "              direct in-line\n";
			std::cout << "---------------------------------------------------------------------\n";
			std::cout << "median time = " << timing2[num_trials/2] * 1e6 << " microseconds\n";
			std::cout << "90% range size = " << (timing2[num_trials*95/100] - timing2[num_trials / 20]) * 1e9 << " nanoseconds\n";
			std::cout << "widest range size (max - min) = " << (timing2[num_trials-1] - timing2[0]) * 1e9 << " nanoseconds\n";
			std::cout << "minimum time = " << timing2[0] * 1e6 << " microseconds\n";
			std::cout << "maximum time = " << timing2[num_trials-1] * 1e6 << " microseconds\n";
			std::cout << "50% range = (" << timing2[num_trials/4] * 1e6 << ", " << timing2[num_trials*3 /4] * 1e6 << ") microseconds\n";
			std::cout << "50% range size = " << (timing2[num_trials*3/4] - timing2[num_trials /4]) * 1e9 << " nanoseconds\n";
			std::cout << "stop the optimiser eliding garbage = " << answer1 << "\n";
			std::cout << "stop the optimiser eliding garbage = " << answer2 << "\n";


			for (size_t j = 0; j< num_trials; ++j )
			{
				timing[j] = timing1[j] - timing2[j];
			}

			std::sort(timing.begin(), timing.end());

			//std::cout << "median time = " << timing[num_trials/2] * 1e6 << " microseconds\n";
			//std::cout << "90% range size = " << (timing[num_trials*95/100] - timing[num_trials / 20]) * 1e9 << " nanoseconds\n";
			//std::cout << "widest range size (max - min) = " << (timing[num_trials-1] - timing[0]) * 1e9 << " nanoseconds\n";
			//std::cout << "minimum time = " << timing[0] * 1e6 << " microseconds\n";
			//std::cout << "maximum time = " << timing[num_trials-1] * 1e6 << " microseconds\n";
			//std::cout << "50% range = (" << timing[num_trials/4] * 1e6 << ", " << timing[num_trials*3 /4] * 1e6 << ") microseconds\n";
			//std::cout << "50% range size = " << (timing[num_trials*3/4] - timing[num_trials /4]) * 1e9 << " nanoseconds\n";
			//std::cout << "stop the optimiser eliding garbage = " << answer1 << "\n";
			//std::cout << "stop the optimiser eliding garbage = " << answer2 << "\n";

			//std::cout << "---------------------------------------------------------------------\n";

			SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);

			return timing[num_trials/2];
	
		}
		static boost::optional<double> overhead_estimate_;
	};

}
#endif
